"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.makeInjectedModule = exports.DIModule = void 0;
var _brandi = require("brandi");
var _function = require("fp-ts/function");
var R = _interopRequireWildcard(require("fp-ts/Reader"));
function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }
function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }
/**
 * Type utils
 */

class DependencyModuleWithTokensImpl extends _brandi.DependencyModule {
  constructor(tokens) {
    super();
    this.tokens = tokens;
    this.allTokens = Object.values(tokens);
  }
}
const bind = tokens => bindings => {
  const mod = new DependencyModuleWithTokensImpl(tokens);
  Object.entries(bindings).filter(([k]) => Object.hasOwn(tokens, k)) //safely use bindings that are mapped to tokens
  .map(([k, v]) => (0, _function.tuple)(k, v)).forEach(([k, v]) => applyScope(v(mod.bind(tokens[k]))));
  return mod;
};
const allBindings = tokens => (0, _function.pipe)(R.ask(), R.local(bind(tokens)));
var Scope = /*#__PURE__*/function (Scope) {
  Scope["CONTAINER"] = "CONTAINER";
  Scope["RESOLUTION"] = "RESOLUTION";
  Scope["TRANSIENT"] = "TRANSIENT";
  Scope["SINGLETON"] = "SINGLETON";
  return Scope;
}(Scope || {});
const applyScope = (syntax, defaultScope = Scope.CONTAINER) => {
  if (!syntax) return;
  const [scopeSyntax, scope] = Array.isArray(syntax) ? syntax : [syntax, defaultScope];
  switch (scope) {
    case Scope.CONTAINER:
      scopeSyntax.inContainerScope();
      break;
    case Scope.RESOLUTION:
      scopeSyntax.inResolutionScope();
      break;
    case Scope.TRANSIENT:
      scopeSyntax.inTransientScope();
      break;
    case Scope.SINGLETON:
      scopeSyntax.inSingletonScope();
      break;
  }
};
const makeInjector = registry => {
  /**
   * Keys that are used to access a token in resulting DependencyModuleWithTokens
   */

  return (target, ...tokensKeys) => {
    registry.set(target, mod => tokensKeys.map(k => mod.tokens[k]));
    return target;
  };
};
class ContravariantMaker {
  constructor(make) {
    this.maker = (0, _function.pipe)(make, R.local(_function.identity));
  }
  make(bindings) {
    return this.maker(bindings); // force call, even if bindings are not exhaustive (assumes "make" was built using previous bindings)
  }
}

// Combine :
// - null intersection : DIModule<T1 & T2, B1 | B2>
// - some intersection : DIModule<T1 & T2, (B1 | B2) \ (B1 & B2)>
// - inclusion : DIModule<T1 & T2, lower(B1 | B2)> (meet semilattice : x <= y when x ^ y = x, greatest lower bound (meet) is with B = never)
// THIS CRAZY TYPE MAKES IT COMMUTATIVE AND IDEMPOTENT !

class DIModule extends ContravariantMaker {
  constructor(injector, make) {
    super(make);
    this.injector = injector;
  }
  bind(key, binding) {
    return new DIModule(this.injector, (0, _function.pipe)(i => this.make(i), R.compose(remainder => ({
      [key]: binding,
      ...remainder
    }))));
  }
  combine(mod2) {
    //TODO: injector may be reused as (callable) class
    const localRegistry = new Map();
    const injector = makeInjector(localRegistry);
    const applyInjector = mod => {
      [...localRegistry.entries()].forEach(([k, v]) => {
        (0, _brandi.injected)(k, ...v(mod));
      });
      return mod;
    };
    return new DIModule(injector, (0, _function.pipe)(R.Do, R.bind('firstMod', () => R.asks(bindings => this.make(bindings))),
    // mod1 receives both PARTIALTOKENS1 and PARTIALTOKENS2, handling potentially unknown tokens is delegated to mod1
    R.bindW('secondMod', () => R.asks(bindings => mod2.make(bindings))), R.bind('bindings', () => R.ask()), R.map(({
      firstMod,
      secondMod,
      bindings
    }) => {
      const tokens = {
        ...firstMod.tokens,
        ...secondMod.tokens
      };

      // compute difference
      const unboundTokens = Object.fromEntries([...Object.entries(tokens)].filter(([k]) => Object.hasOwn(bindings, k)));
      const thirdMod = bind(unboundTokens)(bindings);

      // complete dep module with all tokens. Bindings are used from existing modules.
      const mod = new DependencyModuleWithTokensImpl(tokens);
      firstMod.allTokens.forEach(tok => mod.use(tok).from(firstMod));
      secondMod.allTokens.forEach(tok => mod.use(tok).from(secondMod));
      thirdMod.allTokens.forEach(tok => mod.use(tok).from(thirdMod));
      applyInjector(mod);
      return mod;
    })));
  }
}
exports.DIModule = DIModule;
const makeInjectedModule = tokens => {
  const localRegistry = new Map();
  const injector = makeInjector(localRegistry);
  const applyInjector = (0, _function.pipe)(allBindings(tokens), R.map(mod => {
    [...localRegistry.entries()].forEach(([k, v]) => {
      (0, _brandi.injected)(k, ...v(mod));
    });
    return mod;
  }));
  return new DIModule(injector, applyInjector);
};
exports.makeInjectedModule = makeInjectedModule;
//# sourceMappingURL=brandi.cjs.map