import { DependencyModule, injected, } from 'brandi';
import { identity, pipe, tuple } from 'fp-ts/function';
import * as R from 'fp-ts/Reader';
class DependencyModuleWithTokensImpl extends DependencyModule {
    tokens;
    allTokens;
    constructor(tokens) {
        super();
        this.tokens = tokens;
        this.allTokens = Object.values(tokens);
    }
}
const bind = (tokens) => (bindings) => {
    const mod = new DependencyModuleWithTokensImpl(tokens);
    Object.entries(bindings)
        .filter(([k]) => Object.hasOwn(tokens, k)) //safely use bindings that are mapped to tokens
        .map(([k, v]) => tuple(k, v))
        .forEach(([k, v]) => applyScope(v(mod.bind(tokens[k]))));
    return mod;
};
const allBindings = (tokens) => pipe(R.ask(), R.local(bind(tokens)));
var Scope;
(function (Scope) {
    Scope["CONTAINER"] = "CONTAINER";
    Scope["RESOLUTION"] = "RESOLUTION";
    Scope["TRANSIENT"] = "TRANSIENT";
    Scope["SINGLETON"] = "SINGLETON";
})(Scope || (Scope = {}));
const applyScope = (syntax, defaultScope = Scope.CONTAINER) => {
    if (!syntax)
        return;
    const [scopeSyntax, scope] = Array.isArray(syntax)
        ? syntax
        : [syntax, defaultScope];
    switch (scope) {
        case Scope.CONTAINER:
            scopeSyntax.inContainerScope();
            break;
        case Scope.RESOLUTION:
            scopeSyntax.inResolutionScope();
            break;
        case Scope.TRANSIENT:
            scopeSyntax.inTransientScope();
            break;
        case Scope.SINGLETON:
            scopeSyntax.inSingletonScope();
            break;
    }
};
const makeInjector = (registry) => {
    return (target, ...tokensKeys) => {
        registry.set(target, (mod) => tokensKeys.map((k) => mod.tokens[k]));
        return target;
    };
};
class ContravariantMaker {
    maker;
    constructor(make) {
        this.maker = pipe(make, R.local(identity));
    }
    make(bindings) {
        return this.maker(bindings); // force call, even if bindings are not exhaustive (assumes "make" was built using previous bindings)
    }
}
export class DIModule extends ContravariantMaker {
    injector;
    constructor(injector, make) {
        super(make);
        this.injector = injector;
    }
    bind(key, binding) {
        return new DIModule(this.injector, pipe((i) => this.make(i), R.compose((remainder) => ({
            [key]: binding,
            ...remainder,
        }))));
    }
    combine(mod2) {
        //TODO: injector may be reused as (callable) class
        const localRegistry = new Map();
        const injector = makeInjector(localRegistry);
        const applyInjector = (mod) => {
            [...localRegistry.entries()].forEach(([k, v]) => {
                injected(k, ...v(mod));
            });
            return mod;
        };
        return new DIModule(injector, pipe(R.Do, R.bind('firstMod', () => R.asks((bindings) => this.make(bindings))), // mod1 receives both PARTIALTOKENS1 and PARTIALTOKENS2, handling potentially unknown tokens is delegated to mod1
        R.bindW('secondMod', () => R.asks((bindings) => mod2.make(bindings))), R.bind('bindings', () => R.ask()), R.map(({ firstMod, secondMod, bindings }) => {
            const tokens = {
                ...firstMod.tokens,
                ...secondMod.tokens,
            };
            // compute difference
            const unboundTokens = Object.fromEntries([...Object.entries(tokens)].filter(([k]) => Object.hasOwn(bindings, k)));
            const thirdMod = bind(unboundTokens)(bindings);
            // complete dep module with all tokens. Bindings are used from existing modules.
            const mod = new DependencyModuleWithTokensImpl(tokens);
            firstMod.allTokens.forEach((tok) => mod.use(tok).from(firstMod));
            secondMod.allTokens.forEach((tok) => mod.use(tok).from(secondMod));
            thirdMod.allTokens.forEach((tok) => mod.use(tok).from(thirdMod));
            applyInjector(mod);
            return mod;
        })));
    }
}
export const makeInjectedModule = (tokens) => {
    const localRegistry = new Map();
    const injector = makeInjector(localRegistry);
    const applyInjector = pipe(allBindings(tokens), R.map((mod) => {
        [...localRegistry.entries()].forEach(([k, v]) => {
            injected(k, ...v(mod));
        });
        return mod;
    }));
    return new DIModule(injector, applyInjector);
};
//# sourceMappingURL=brandi.js.map